#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <termios.h>
#include <dirent.h>

#define MIN_LENGTH 32

typedef unsigned char ch;
typedef struct {unsigned t[0x100]; ch *s;} Table;
typedef struct {unsigned length,max_length; ch *s;} Buffer;

/* Global variables */
Table t,t1,t2;
Buffer in_name,out_name;
ch *arg,*path;
int readt,written,inter;
FILE *in; FILE *out;
struct termios saved_attributes;

/* Buffer handling */
void BufferInit(Buffer *buf)
{ buf->length=0; buf->max_length=MIN_LENGTH; buf->s=malloc(MIN_LENGTH); }

void BufferFin(Buffer *buf)
{ buf->length=0; if (buf->s) free(buf->s); }

void BufferPut(Buffer *buf, ch c)
{ if (buf->length>=buf->max_length)
  { ch *s; buf->max_length*=2; s=malloc(buf->max_length);
    memcpy(s,buf->s,buf->length); free(buf->s); buf->s=s;
  }
  buf->s[(buf->length)++]=c;
}

/* Write error message */
void Warning(ch *s1, ch *s2)
{ fprintf(stderr,"%s %s\n",s1,s2); }

/* Write error message and die */
void Error(ch *s1, ch *s2)
{ fprintf(stderr,"%s %s\n",s1,s2); exit(1); }

/*  0 = does not exist
    1 = file
    2 = directory
 */
int ftype(char *name)
{ struct stat s;
  if (stat(name,&s)) return(0);
  if (S_ISDIR(s.st_mode)) return(2);
  return(1);
}

/* Initialize t if nothing read yet */
void Init(void)
{ unsigned i;
  if (!readt)
  { for (i=0; i<0x100; i++) t.t[i]=i;
    t.s=NULL; readt=1;
  }
}

/* Read the code table t from the file in the binary format */
void ReadBin(Table *t)
{ unsigned i; int c;
  FILE *in=fopen(in_name.s,"r");
  if (!in) Error("Cannot open",in_name.s);
  /* main part of the code table */
  for (i=0; i<0x100; i++)
  { c=fgetc(in);
    if (c<0) Error("Unexpected end of",in_name.s);
    t->t[i]=(unsigned)c;
  }
  /* tail of the code table */
  if (t->s) free(t->s); t->s=NULL;
  c=fgetc(in);
  if (c>=0)
  { Buffer b; BufferInit(&b);
    do
    { t->t[c]=b.length+0x100;
      do
      { c=fgetc(in);
        if (c<0) Error("Unexpected end of",in_name.s);
        BufferPut(&b,(ch)c);
        if (c==1)
        { c=fgetc(in);
          if (c<0) Error("Unexpected end of",in_name.s);
          BufferPut(&b,(ch)c); c=1;
        }
      }
      while (c);
      c=fgetc(in);
    }
    while (c>=0);
    t->s=malloc(b.length); memcpy(t->s,b.s,b.length); BufferFin(&b);
  }
  fclose(in);
}
 
/* Read the code table t from the file in the text format */
void ReadText(Table *t)
{ int i,j,c; Buffer b1,b2,b3; int ct[0x101];
  FILE *in=fopen(in_name.s,"r");
  if (!in) Error("Cannot open",in_name.s);
  for (i=0; i<0x100; i++) t->t[i]=i;
  /* char types */
  for (i=1; i<0x101; i++) ct[i]=-1;       /* nothing special */
  for (i=0x31; i<0x3B; i++) ct[i]=i-0x31; /* 0..9 */
  for (i=0x42; i<0x48; i++) ct[i]=i-0x38; /* A..F */
  for (i=0x62; i<0x68; i++) ct[i]=i-0x58; /* a..f */
  ct[0x24]=-2; ct[0x23]=-3; ct[0x28]=-3;  /* # " ' */
  ct[0x0B]=-4; ct[0]=-5;                  /* endline endfile */
  BufferInit(&b1); BufferInit(&b2); BufferInit(&b3);
  /* main loop */
  do
  { b2.length=0; c=fgetc(in); i=ct[c+1];
    while (1)
    { if (i>=0) /* hex digit */
      { b1.length=0;
        do { BufferPut(&b1,(ch)i); c=fgetc(in); i=ct[c+1]; }
        while (i>=0);
        if (b1.length&1) { BufferPut(&b2,b1.s[0]); j=1; }
        else { j=0; }
        for (; j<b1.length; j+=2)
          BufferPut(&b2,(ch)(0x10*b1.s[j]+b1.s[j+1]));
      }
      else if (i==-3) /* string */
      { j=c;
        while (1)
        { c=fgetc(in);
          if ((ct[c+1]<-3)||(c==j)) break;
          BufferPut(&b2,(ch)c);
        }
        if (c==j) { c=fgetc(in); i=ct[c+1]; }
      }
      else if (i==-2) /* comment */
      { do { c=fgetc(in); i=ct[c+1]; } while (i>=-3);
        break;
      }
      else if (i<-3) /* endline or endfile */
      { break; }
      else { c=fgetc(in); i=ct[c+1]; }
    }
    /* putting the line to the table */
    if (b2.length==2) { t->t[b2.s[0]]=b2.s[1]; }
    else if (b2.length>0)
    { t->t[b2.s[0]]=b3.length+0x100;
      for (j=1; j<b2.length; j++)
      { if (b2.s[j]<=1) BufferPut(&b3,(ch)1);
        BufferPut(&b3,b2.s[j]);
      }
      BufferPut(&b3,(ch)0);
    }
  }
  while (i>-5);
  if (t->s) free(t->s);
  if (b3.length) { t->s=malloc(b3.length); memcpy(t->s,b3.s,b3.length); }
  else t->s=NULL;
  BufferFin(&b1); BufferFin(&b2); BufferFin(&b3);
}

/* Invert the code table t1 to t2 */
void Invert(Table *t1, Table *t2)
{ unsigned i,j;
  if (t2->s) free(t2->s); t2->s=NULL;
  for (i=0; i<0x100; i++)
  { t2->t[i]=i; j=0;
    while (1)
    { if ((t1->t[j]==i)&&(j!=i)) { t2->t[i]=j; break; }
      else if ((j++)>=0xFF) break;
    }
  }
}

/* Multiply the code table t1 by t2 */
void Mult(Table *t1, Table *t2)
{ Table t; Buffer b1,b2,b3; unsigned i,j,k; ch b;
  BufferInit(&b1); BufferInit(&b2); BufferInit(&b3);
  for (i=0; i<0x100; i++)
  { b1.length=0; b2.length=0;
    if ((j=t1->t[i])<0x100) BufferPut(&b1,(ch)j);
    else
    { j-=0x100;
      while (1)
      { if (!(b=t1->s[j])) break;
        j++;
        if (b==(ch)1) b=t1->s[j++];
        BufferPut(&b1,b);
      }
    }
    for (j=0; j<b1.length; j++)
    { if ((k=t2->t[b1.s[j]])<0x100) BufferPut(&b2,(ch)k);
      else
      { k-=0x100;
        while (1)
        { if (!(b=t2->s[k])) break;
          k++;
          if (b==1) b=t2->s[k++];
          BufferPut(&b2,(ch)b);
        }
      }
    }
    if (b2.length==1) t.t[i]=b2.s[0];
    else
    { t.t[i]=b3.length+0x100;
      for (j=0; j<b2.length; j++)
      { if (b2.s[j]<=1) BufferPut(&b3,(ch)1);
        BufferPut(&b3,b2.s[j]);
      }
      BufferPut(&b3,(ch)0);
    }
  }
  if (t1->s) free(t1->s);
  if (b3.length) { t1->s=malloc(b3.length); memcpy(t1->s,b3.s,b3.length); }
  else t1->s=NULL;
  memcpy(&t1->t,&t.t,sizeof(t.t));
  BufferFin(&b1); BufferFin(&b2); BufferFin(&b3);
}

/* Read the code table from the file */
void Read1(int bin, int inv)
{ if (bin)
  { if (readt)
    { if (inv) { ReadBin(&t2); Invert(&t2,&t1); } else ReadBin(&t1);
      Mult(&t,&t1);
    }
    else
    { if (inv) { ReadBin(&t2); Invert(&t2,&t); } else ReadBin(&t);
      readt=1;
    }
  }
  else
  { if (readt)
    { if (inv) { ReadText(&t2); Invert(&t2,&t1); } else ReadText(&t1);
      Mult(&t,&t1);
    }
    else
    { if (inv) { ReadText(&t2); Invert(&t2,&t); } else ReadText(&t);
      readt=1;
    }
  }
}

/* Read the code table */
void Read(int inv)
{ unsigned i,j; ch c;
  if (!arg[1])
  { inter=1; return; }
  else if (arg[1]!='/')
  { i=0;
    do
    { in_name.length=0;
      while (1)
      { c=path[i++];
        if ((!c)||(c==':')) break;
        BufferPut(&in_name,c);
      }
      if (!c) i=-1;
      if ((in_name.length>0)&&(in_name.s[in_name.length-1]=='/'));
      else BufferPut(&in_name,'/');
      j=1;
      do BufferPut(&in_name,c=arg[j++]); while (c);
      /* try path/arg */
      if (ftype(in_name.s)==1) { Read1(1,inv); return; }
      /* try path/arg.16 */
      in_name.length--;
      BufferPut(&in_name,'.'); BufferPut(&in_name,'1'); BufferPut(&in_name,'6'); BufferPut(&in_name,(ch)0);
      if (ftype(in_name.s)==1) { Read1(0,inv); return; }
    }
    while (i!=-1);
  }
  else
  { j=1;
    do BufferPut(&in_name,c=arg[j++]); while (c);
    /* try arg */
    if (ftype(in_name.s)==1) { Read1(1,inv); return; }
    /* try arg.16 */
    in_name.length--;
    BufferPut(&in_name,'.'); BufferPut(&in_name,'1'); BufferPut(&in_name,'6'); BufferPut(&in_name,(ch)0);
    if (ftype(in_name.s)==1) { Read1(0,inv); return; }
  }
  Error("Cannot open",arg+1);
}

/* Write the code table to a text file */
void WriteText(FILE *out)
{ int i,j; ch b;
  for (i=0; i<0x100; i++) if ((j=t.t[i])!=i)
  { fprintf(out,"%02x",i);
    if (j<0x100) fprintf(out," %02x",j);
    else
    { j-=0x100;
      while (1)
      { b=t.s[j];
        if (!b) break; j++;
        if (b==(ch)1) b=t.s[j++];
        fprintf(out," %02x",(int)b);
      }
    }
    fputc(0xA,out);
  }
}

/* Write the code table to a binary file */
void WriteBin(FILE *out)
{ int i,j; ch b;
  /* main part of the code table */
  for (i=0; i<0x100; i++)
  { if ((j=t.t[i])<0x100) fputc(j,out); else fputc(0,out); }
  /* tail of the code table */
  for (i=0; i<0x100; i++) if ((j=t.t[i])>=0x100)
  { fputc(i,out); j-=0x100;
    while (1)
    { b=t.s[j++]; fputc((int)b,out);
      if (b==(ch)1) fputc((int)t.s[j++],out);
      else if (!b) break;
    }
  }
}

/* Write the code table to the file */
void Write(void)
{ int i; FILE *out;
  Init(); written=1;
  if (!arg[1]) { WriteText(stdout); return; }
  if (!(out=fopen(++arg,"w"))) { Warning("Cannot open",arg); return; }
  i=strlen(arg)-3;
  if ((i>0)&&(arg[i]=='.')&&(arg[i+1]=='1')&&(arg[i+2]=='6'))
    WriteText(out);
  else WriteBin(out);
  fclose(out);
}

/* Copy in to out using the code table */
void Do(FILE *in, FILE *out)
{ unsigned i; int c; ch b;
  Init();
  while (1)
  { c=fgetc(in); if (c<0) break; i=t.t[c];
    if (i<0x100) fputc(i,out);
    else
    { i-=0x100;
      while (1)
      { b=t.s[i];
        if (!b) break; i++;
        if (b==(ch)1) b=t.s[i++];
        fputc((int)b,out);
      }
    }
  }
  fclose(in); fclose(out);
}

/* Reset terminal input mode */
void reset_mode(void)
{ tcsetattr(STDIN_FILENO,TCSANOW,&saved_attributes); }

/* Copy terminal input to stdout using the code table */
void Inter(void)
{ unsigned i; ch b; int tty;
  struct termios attributes;
  Init();
  if (tty=isatty(STDIN_FILENO))
  { /* Save the terminal attributes so we can restore them later */
    tcgetattr(STDIN_FILENO,&saved_attributes); atexit(reset_mode);
    /* Set the raw terminal mode */
    tcgetattr(STDIN_FILENO,&attributes);
    /*
    attributes.c_lflag&=~(ICANON|ECHO|ISIG);
    attributes.c_cc[VMIN]=1; attributes.c_cc[VTIME]=0;
    */
    cfmakeraw(&attributes);
    tcsetattr(STDIN_FILENO,TCSAFLUSH,&attributes);
  }
  while (1)
  { if (!read(STDIN_FILENO,&b,1)) break;
    /* if (tty&&(b==0x4)) break; */
    i=t.t[(int)b];
    if (i<0x100)
    { b=(ch)i; write(STDOUT_FILENO,&b,1); }
    else
    { i-=0x100;
      while (1)
      { b=t.s[i];
        if (!b) break; i++;
        if (b==(ch)1) b=t.s[i++];
        write(STDOUT_FILENO,&b,1);
      }
    }
  }
  if (tty) reset_mode();
}

void Dir(char *n);

/* Recursively copy the directory in_name to the directory out_name */
void RecDir()
{ DIR *dir; struct dirent *ent; int i,n,l=in_name.length; ch c; char *name;
  switch (ftype(out_name.s))
  { case 1: Warning("Cannot copy to",out_name.s); return;
    case 0: mkdir(out_name.s,0777);
    case 2: ;
  }
  if (!(dir=opendir(in_name.s)))
  { Warning("Cannot open directory",in_name.s); return; }
  in_name.s[in_name.length-1]='/';
  while (ent=readdir(dir))
  { if (((n=strlen(name=ent->d_name))==1) && (name[0]=='.')) continue;
    if ((n==2) && (name[0]=='.') && (name[1]=='.')) continue;
    for (i=0; c=name[i]; i++) BufferPut(&in_name,c);
    BufferPut(&in_name,(ch)0);
    Dir(in_name.s);
    in_name.length=l;
  }
}

/* Copy the file n to the directory out_name */
void Dir(char *n)
{ int i,l=out_name.length; ch c;
  /* Constructing the output file name */
  out_name.s[out_name.length-1]='/';
  i=strlen(n)-2;
  while ((i>=0)&&(n[i]!='/')) i--;
  for (i++; c=n[i]; i++) BufferPut(&out_name,c);
  BufferPut(&out_name,(ch)0);
  /* Copying */
  switch (ftype(n))
  { case 0: Warning(n,"does not exist"); break;
    case 1:
      if (!(in=fopen(n,"r"))) Error("Cannot open",n);
      if (!(out=fopen(out_name.s,"w"))) Error("Cannot open",out_name.s);
      Do(in,out);
      break;
    case 2:
      in_name.length=0;
      for (i=0; c=n[i]; i++) BufferPut(&in_name,c);
      if (in_name.s[in_name.length-1]=='/') in_name.length--;
      BufferPut(&in_name,(ch)0);
      RecDir();
  }
  out_name.length=l;
}

/* Main program */
int main(int argc, char **argv)
{ int i,n,ft; ch *arg0; ch c;
  /* obtaining path */
  if (!(path=getenv("CCPPATH"))) path=DEFAULT;
  readt=0; written=0; inter=0; n=0;
  /* processing options */
  t.s=NULL; t1.s=NULL; t2.s=NULL;
  BufferInit(&in_name); BufferInit(&out_name);
  for (i=1; i<argc; i++)
  { arg=argv[i];
    if (arg[0]=='-') Read(0);
    else if (arg[0]=='+') Read(1);
    else if (arg[0]=='=') Write();
    else argv[n++]=arg;
  }
  /* processing args */
  if (inter)
  { if (n) Error("No file arguments are allowed in the \"-\" mode","");
    Inter();
  }
  else if (!n)
  { if (!written) Do(stdin,stdout); }
  else if (n==1)
  { if (ftype(arg=argv[0])!=1) Error(arg,"is not a file");
    if (!(in=fopen(arg,"r"))) Error("Cannot open",argv[0]);
    Do(in,stdout);
  }
  else if (ft=ftype(arg=argv[n-1])==2) /* last arg is a directory */
  { out_name.length=0;
    for (i=0; c=arg[i]; i++) BufferPut(&out_name,c);
    if (out_name.s[out_name.length-1]=='/') out_name.length--;
    BufferPut(&out_name,(ch)0);
    for (i=0; i<n-1; i++) Dir(argv[i]);
  }
  else if (n>2) Error("Usage: ccp [options] [file [file]] |",
        "ccp [options] files directory");
  else if (ft==1) /* last arg is a file */
  { if (ftype(arg0=argv[0])!=1) Error(arg0,"is not a file");
    if (!(in=fopen(arg0,"r"))) Error("Cannot open",arg0);
    if (!(out=fopen(arg,"w"))) Error("Cannot open",arg);
    Do(in,out);
  }
  else /* last arg does not exist, n==2 */
    switch (ftype(arg0=argv[0]))
    { case 0: Error(arg0,"does not exist");
      case 1:
        if (!(in=fopen(arg0,"r"))) Error("Cannot open",arg0);
        if (!(out=fopen(arg,"w"))) Error("Cannot open",arg);
        Do(in,out);
        break;
      case 2:
        out_name.length=0;
        for (i=0; c=arg[i]; i++) BufferPut(&out_name,c);
        if (out_name.s[out_name.length-1]=='/') out_name.length--;
        BufferPut(&out_name,(ch)0);
        in_name.length=0;
        for (i=0; c=arg0[i]; i++) BufferPut(&in_name,c);
        if (in_name.s[in_name.length-1]=='/') in_name.length--;
        BufferPut(&in_name,(ch)0);
        RecDir();
    }
  exit(0);
}
